@@ -1,5 +1,4 @@
-#!/usr/local/bin/python3.6
-# vim: expandtab shiftwidth=4
+#!/usr/bin/env python3
 
 # Copyright (C) 2019 National Institute of Informatics
 #
@@ -26,6 +25,10 @@ class SinetError(Exception):
     pass
 
 
+class InvalidConfigError(SinetError):
+    pass
+
+
 class NoServiceError(SinetError):
     pass
 
